@@ -1,5 +1,5 @@
 /* Copyright (c) 2009, 2013, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2020, MariaDB Corporation.
+   Copyright (c) 2009, 2022, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -201,13 +201,12 @@ void debug_sync_end(void)
 
     /* Print statistics. */
     {
-      char llbuff[22];
-      sql_print_information("Debug sync points hit:                   %22s",
-                            llstr(debug_sync_global.dsp_hits, llbuff));
-      sql_print_information("Debug sync points executed:              %22s",
-                            llstr(debug_sync_global.dsp_executed, llbuff));
-      sql_print_information("Debug sync points max active per thread: %22s",
-                            llstr(debug_sync_global.dsp_max_active, llbuff));
+      sql_print_information("Debug sync points hit:                   %lld",
+                            debug_sync_global.dsp_hits);
+      sql_print_information("Debug sync points executed:              %lld",
+                            debug_sync_global.dsp_executed);
+      sql_print_information("Debug sync points max active per thread: %lld",
+                            debug_sync_global.dsp_max_active);
     }
   }
 
@@ -348,12 +347,13 @@ static char *debug_sync_bmove_len(char *to, char *to_end,
   DBUG_ASSERT(to_end);
   DBUG_ASSERT(!length || from);
   set_if_smaller(length, (size_t) (to_end - to));
-  memcpy(to, from, length);
+  if (length)
+    memcpy(to, from, length);
   return (to + length);
 }
 
 
-#if !defined(DBUG_OFF)
+#ifdef DBUG_TRACE
 
 /**
   Create a string that describes an action.
@@ -443,8 +443,7 @@ static void debug_sync_print_actions(THD *thd)
 
   DBUG_VOID_RETURN;
 }
-
-#endif /* !defined(DBUG_OFF) */
+#endif /* defined(DBUG_TRACE) */
 
 
 /**
@@ -775,6 +774,7 @@ static bool debug_sync_set_action(THD *thd, st_debug_sync_action *action)
   else
   {
     const char *dsp_name= action->sync_point.c_ptr();
+#ifdef DBUG_TRACE
     DBUG_EXECUTE("debug_sync", {
         /* Functions as DBUG_PRINT args can change keyword and line nr. */
         const char *sig_emit= action->signal.c_ptr();
@@ -785,6 +785,7 @@ static bool debug_sync_set_action(THD *thd, st_debug_sync_action *action)
                     dsp_name, action->activation_count,
                     action->hit_limit, action->execute, action->timeout,
                     sig_emit, sig_wait));});
+#endif
 
     /* Check this before sorting the array. action may move. */
     is_dsp_now= !my_strcasecmp(system_charset_info, dsp_name, "now");
@@ -797,7 +798,9 @@ static bool debug_sync_set_action(THD *thd, st_debug_sync_action *action)
                sizeof(st_debug_sync_action), debug_sync_qsort_cmp);
     }
   }
+#ifdef DBUG_TRACE
   DBUG_EXECUTE("debug_sync_list", debug_sync_print_actions(thd););
+#endif
 
   /* Execute the special sync point 'now' if activated above. */
   if (is_dsp_now)
@@ -995,6 +998,15 @@ static char *debug_sync_number(ulong *number_p, char *actstrptr,
     The input string needs to be ASCII NUL ('\0') terminated. We split
     nul-terminated tokens in it without copy.
 
+  @note
+    The current implementation does not support two 'now SIGNAL xxx' commands
+    in a row for multiple threads as the first one can get lost while
+    the waiting threads are sleeping on mysql_cond_timedwait().
+    One reason for this is that the signal name is stored in a global variable
+    that is overwritten.  A better way would be to store all signals in
+    an array together with a 'time' when the signal was sent. This array
+    should be checked on broadcast.
+
   @see the function comment of debug_sync_token() for more constraints
     for the string.
 */
@@ -1358,7 +1370,7 @@ uchar *debug_sync_value_ptr(THD *thd)
 
 static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
 {
-#ifndef DBUG_OFF
+#ifdef DBUG_TRACE
   const char *dsp_name= action->sync_point.c_ptr();
   const char *sig_emit= action->signal.c_ptr();
   const char *sig_wait= action->wait_for.c_ptr();
@@ -1449,12 +1461,12 @@ static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
         restore_current_mutex = false;
 
       set_timespec(abstime, action->timeout);
-      DBUG_EXECUTE("debug_sync_exec", {
+      DBUG_EXECUTE("debug_sync_exec",
           /* Functions as DBUG_PRINT args can change keyword and line nr. */
-          const char *sig_glob= debug_sync_global.ds_signal.c_ptr();
           DBUG_PRINT("debug_sync_exec",
                      ("wait for '%s'  at: '%s'  curr: '%s'",
-                      sig_wait, dsp_name, sig_glob));});
+                      sig_wait, dsp_name,
+                      debug_sync_global.ds_signal.c_ptr())););
 
       /*
         Wait until global signal string matches the wait_for string.
@@ -1463,17 +1475,17 @@ static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
         the required dynamic memory allocated.
       */
       while (stringcmp(&debug_sync_global.ds_signal, &action->wait_for) &&
-             !thd->killed && opt_debug_sync_timeout)
+             !(thd->killed & KILL_HARD_BIT) && opt_debug_sync_timeout)
       {
         error= mysql_cond_timedwait(&debug_sync_global.ds_cond,
                                     &debug_sync_global.ds_mutex,
                                     &abstime);
-        DBUG_EXECUTE("debug_sync", {
+        DBUG_EXECUTE("debug_sync",
             /* Functions as DBUG_PRINT args can change keyword and line nr. */
-            const char *sig_glob= debug_sync_global.ds_signal.c_ptr();
             DBUG_PRINT("debug_sync",
                        ("awoke from %s  global: %s  error: %d",
-                        sig_wait, sig_glob, error));});
+                        sig_wait, debug_sync_global.ds_signal.c_ptr(),
+                        error)););
         if (unlikely(error == ETIMEDOUT || error == ETIME))
         {
           // We should not make the statement fail, even if in strict mode.
